import logging
from pathlib import Path

import yaml
from inspect_ai.dataset import Sample
from pydantic import ValidationError

from inspect_cyber.dataset import AgenticEvalDataset
from inspect_cyber.model import Eval, Sandbox

logger = logging.getLogger(__name__)

POSSIBLE_EVAL_FILE_NAMES = ["eval.yaml", "eval.yml"]


def create_agentic_eval_dataset(
    root_dir: str | Path,
    name: str | None = None,
) -> AgenticEvalDataset:
    """
    Create an AgenticEvalDataset from the given root_dir.

    Args:
        root_dir (str | Path): Directory used to discover evals. Must be an
            absolute path.
        name (str | None): Optonal name of the dataset.

    Returns:
        AgenticEvalDataset: A dataset containing samples for the evals and
            variants found in the root_dir.
    """
    if isinstance(root_dir, str):
        root_dir = Path(root_dir)
    _validate_root_dir(root_dir)

    eval_file_paths = _get_eval_file_paths(root_dir)

    eval_directories = set()
    for path in eval_file_paths:
        directory = path.parent
        if directory in eval_directories:
            raise ValueError(
                f"Multiple eval files found in the same directory: {directory}. "
                "Please ensure that each eval file is in a separate directory."
            )
        eval_directories.add(directory)

    samples = _create_samples(eval_file_paths)
    return AgenticEvalDataset(root_dir, samples, name)


def _validate_root_dir(root_dir: Path) -> None:
    if not root_dir.is_absolute():
        raise ValueError(f"root_dir must be an absolute path. Got '{root_dir}'")
    if not root_dir.exists():
        raise ValueError(
            f"The path '{root_dir}' does not exist. Please ensure the path is correct "
            f"and the directory has been created."
        )
    if not root_dir.is_dir():
        raise ValueError(f"The path '{root_dir}' is not a directory.")


def _get_eval_file_paths(root_dir: Path) -> list[Path]:
    paths = [path for p in POSSIBLE_EVAL_FILE_NAMES for path in root_dir.rglob(p)]

    def is_hidden(path: Path):
        return any(part.startswith(".") for part in path.relative_to(root_dir).parts)

    paths = [path for path in paths if not is_hidden(path)]
    return paths


def _create_samples(
    eval_file_paths: list[Path],
) -> list[Sample]:
    """Create samples for each variant in the eval files provided."""
    samples = []
    for eval_file_path in eval_file_paths:
        eval = _load_eval(eval_file_path)
        eval_sandbox = (
            _resolve_sandbox(eval.sandbox, eval_file_path.parent)
            if eval.sandbox
            else None
        )
        eval_files = _resolve_paths(eval.files, eval_file_path.parent)
        eval_setup = (
            _resolve_path(eval.setup, eval_file_path.parent) if eval.setup else None
        )

        for variant_name, variant in eval.variants.items():
            variant_sandbox = (
                _resolve_sandbox(variant.sandbox, eval_file_path.parent)
                if variant.sandbox
                else None
            )
            variant_files = _resolve_paths(variant.files, eval_file_path.parent)
            variant_setup = (
                _resolve_path(variant.setup, eval_file_path.parent)
                if variant.setup
                else None
            )

            sample = Sample(
                id=f"{eval.name} ({variant_name})",
                input=variant.prompt,
                target=variant.flag or eval.flag or "",
                sandbox=variant_sandbox or eval_sandbox,
                files=eval_files | variant_files,
                setup=variant_setup or eval_setup,
                metadata={
                    "eval_name": eval.name,
                    "eval_file_path": eval_file_path,
                    "variant_name": variant_name,
                    **eval.metadata,  # Eval metadata overrides base metadata
                    **variant.metadata,  # Variant metadata overrides eval metadata
                },
            )
            samples.append(sample)

    return samples


def _load_eval(eval_file_path: Path) -> Eval:
    with open(eval_file_path, "r") as f:
        data = yaml.safe_load(f)
        try:
            return Eval(**data)
        except ValidationError:
            logger.error("Unable to process %s", eval_file_path)
            raise


def _resolve_paths(files: dict[str, str], base_path: Path) -> dict[str, str]:
    return {
        file_path_in_sandbox: _resolve_path(local_file_path, base_path)
        for file_path_in_sandbox, local_file_path in files.items()
    }


def _resolve_path(path: str, base_path: Path) -> str:
    path_obj = Path(path)
    full_path = path_obj if path_obj.is_absolute() else base_path / path

    if not full_path.exists():
        raise ValueError(
            f"Path not found: {full_path}. Please check that the path exists "
            "and you have permission to access it."
        )

    return str(full_path.resolve())


def _resolve_sandbox(sandbox: Sandbox, base_path: Path) -> tuple[str, str]:
    config_file = base_path / sandbox.config

    if not config_file.exists():
        raise FileNotFoundError(
            f"Configuration file for Sandbox({sandbox.type}, {sandbox.config}) not "
            f"found at {config_file}. Please check that the file exists and "
            "you have permission to access it."
        )

    return sandbox.type, str(config_file.resolve())
