from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field, field_validator, model_validator


class Eval(BaseModel, frozen=True):
    """Represents the contents of an eval.yaml file."""

    name: str
    """Name of the eval."""

    sandbox: Sandbox | None = Field(default=None)
    """Sandbox environment configuration."""

    files: dict[str, str] = Field(default_factory=dict)
    """Files to copy into the sandbox environment(s)."""

    setup: str | None = Field(default=None)
    """Setup script to run within the 'default' sandbox environment."""

    flag: str | None = Field(default=None)
    """Flag to use for scoring."""

    metadata: dict = Field(default_factory=dict)
    """Arbitrary metadata to associate with the eval."""

    variants: dict[str, Variant]
    """Variants of the eval."""

    @field_validator("variants")
    def validate_variants(cls, variants: dict[str, Variant]) -> dict[str, Variant]:
        if len(variants) < 1:
            raise ValueError("At least one variant is required.")
        return variants

    @model_validator(mode="after")
    def validate_sandbox_exists(self) -> "Eval":
        """Ensure each variant has a sandbox (either from the eval or itself)."""
        for variant_name, variant in self.variants.items():
            if variant.sandbox is None and self.sandbox is None:
                raise ValueError(
                    "Sandbox must be specified either in "
                    f"the eval '{self.name}' or in its variant '{variant_name}'"
                )
        return self


class Variant(BaseModel, frozen=True):
    """Represents a variant of an Eval."""

    prompt: str
    """Prompt for the variant."""

    sandbox: Sandbox | None = None
    """Sandbox environment configuration.
    If specified, takes precedence over the Eval's sandbox."""

    files: dict[str, str] = Field(default_factory=dict)
    """Additional files to copy into the sandbox environment(s).
    If the same file path for the same sandbox environment is specified in both the
    Eval and Variant, the Variant's file takes precedence."""

    setup: str | None = None
    """Setup script to run within the 'default' sandbox environment.
    If specified, takes precedence over the Eval's setup script."""

    flag: str | None = None
    """Flag to use for scoring.
    If specified, takes precedence over the Eval's flag."""

    metadata: dict = Field(default_factory=dict)
    """Arbitrary metadata to associate with the variant."""


class Sandbox(BaseModel, frozen=True):
    """Represents a sandbox environment configuration."""

    type: str
    """The type of sandbox environment."""

    config: str
    """Path to the config file for the sandbox environment."""


AgenticEvalMetadataKey = Literal["eval_name", "variant_name"] | str
AgenticEvalMetadata = dict[AgenticEvalMetadataKey, str]
