from pathlib import Path
from typing import Any

from inspect_ai import Task, task
from inspect_ai.scorer import includes
from inspect_ai.solver import basic_agent
from inspect_ai.tool import bash, python

from inspect_cyber.dataset import create_agentic_eval_dataset


@task
def cyber_task(
    root_dir: str | None = None,
    eval_names: str | list[str] | None = None,
    variant_names: str | list[str] | None = None,
    metadata_filters: str | list[str] | dict[str, Any] | None = None,
    sandboxes: str | list[str] | None = None,
) -> Task:
    """
    Create a task for the Cyber Agentic Evaluation dataset.

    Args:
        root_dir (str | None): The root directory of the dataset. If None, uses the current working directory.
        eval_names (str | list[str] | None): The names of the evaluations to include. If None, includes all.
        variant_names (str | list[str] | None): The names of the variants to include. If None, includes all.
        metadata_filters (dict[str, Any] | None): A dictionary of metadata filters to apply. If None, no filters are applied.
        sandboxes (str | list[str] | None): The names of the sandboxes to include. If None, includes all.

    Returns:
        Task: The task object.
    """  # noqa: E501
    if root_dir is None:
        root_dir_path = Path.cwd()
    else:
        root_dir_path = Path(root_dir)

    if isinstance(metadata_filters, str):
        metadata_filters = parse_metadata_filters(metadata_filters.split(","))
    elif isinstance(metadata_filters, list):
        metadata_filters = parse_metadata_filters(metadata_filters)

    dataset = (
        create_agentic_eval_dataset(root_dir_path.absolute())
        .filter_by_metadata_field("eval_name", eval_names)
        .filter_by_metadata_field("variant_name", variant_names)
        .filter_by_metadata(metadata_filters)
        .filter_by_sandbox(sandboxes)
    )
    return Task(
        dataset=dataset,
        solver=basic_agent(tools=[bash(), python()]),
        name=dataset.name,
        scorer=includes(),
    )


def parse_metadata_filters(metadata_filters: list[str]) -> dict[str, str]:
    """
    Parse metadata filters from key=value strings to a dictionary.

    Args:
        metadata_filters: List of metadata filters in "key=value" format.
            Example: ["key1=value1", "key2=value2"]

    Returns:
        Dictionary mapping metadata keys to their values.
    """
    filters = {}
    for filter_str in metadata_filters:
        if "=" not in filter_str:
            raise ValueError(
                f"Invalid metadata filter: {filter_str}. Expected 'key=value' format."
            )

        key, value = filter_str.split("=", 1)
        filters[key] = value

    return filters
